/*
 * Picture Puzzle Collection
 * Copyright (c) 1998-1999,2003-2006 Blazing Games Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.blazinggames.pp;

import java.awt.*;
import java.awt.event.*;
import java.util.Random;

/**
 * The Picture Puzzle game screen
 * 
 * @author Billy D. Spelchan
 * @version 1.0 OSR
 * CREATED: December 1998
*/
public class PicturePuzzle
extends Panel
implements PuzScreen
{
// *********************
// ***** CONSTANTS *****
// *********************
	private static final long serialVersionUID = 8080710100L;
	
    private static final String TIME_LABEL = "TIME";
    private static final String DONE_LABEL = "% DONE";
    private static final String MOVES_LABEL = "MOVES";
    private static final Font LABEL_FONT = new Font("Serif", 0, 40);


// *********************
// ***** VARIABLES *****
// *********************
    
  /** The image for the puzzle */
  Image iPuzzle;

  /** Size of the grid*/
  int nGridWidth, nGridHeight;

  /** Size of each cell of the grid */
  int nGridPixelsX, nGridPixelsY;

  /** Border size of image (for off-sized images) */
  int nPicTopGap, nPicBotGap, nPicLeftGap, nPicRightGap;

  /** Color of grid */
  Color colorCorrect, colorIncorrect, colorSelect;

  /** The actual puzzle layout data */
  int nPuzzleData[];

  /** Currently Selected puzzle piece.  -1 for none! */
  int nCurrentSelection, nSecondSelection;

  /** animation timing */
  int nAnimationFrame, nAnimationRate;

  /** game states */
  boolean hasWon, animationPlaying, bGridVisible;

  protected Rectangle _playfieldRect;
  protected boolean _usePlayfield;
  protected Rectangle _counterRect;
  protected int _counterValue;
  protected boolean _counterVisible;
  protected Rectangle _timerRect;
  protected int _timerValue;
  protected boolean _timerVisible;
  protected Rectangle _percentRect;
  protected int _percentValue;
  protected boolean _percentVisible;
  protected long _startTime;
  
// *** CONSTRUCTOR ***
	PicturePuzzle(MouseListener ml) {
		addMouseListener(ml);

		setBackground(Color.white);
        colorCorrect = Color.green;
        colorIncorrect = Color.red;
        colorSelect = Color.blue;
        bGridVisible = true;

        _counterVisible = false;
        _timerVisible = false;
        _percentVisible = false;
    }

  /**
    Sets the image and piece size for the puzzle.  The image must be loaded
    into memory before it is used due to the way the image class works.  The
    gpx and gpy variables are the width and height of the piece in pixels.
  */
  public void setPuzzle(Image i, int gpx, int gpy) {
    int temp;

    iPuzzle = i;
    nGridPixelsX = gpx;
    nGridPixelsY = gpy;

    // calculate width and left/right borders
    nGridWidth = i.getWidth(this) / gpx;
    temp = i.getWidth(this) - nGridWidth * gpx;
    nPicLeftGap = temp / 2;
    nPicRightGap = temp - nPicLeftGap;

    // calculate height and top/bottom borders
    nGridHeight = i.getHeight(this) / gpy;
    temp = i.getHeight(this) - nGridHeight * gpy;
    nPicTopGap = temp / 2;
    nPicBotGap = temp - nPicTopGap;

    // Initialize the puzzle grid
    temp = nGridHeight * nGridWidth;
    nPuzzleData = new int[temp];
    for (int cntr = 0; cntr < temp; ++cntr)
      nPuzzleData[cntr] = cntr;
  }

  /**
    Controls the speed of the animation.  Setting the speed to 0 disables
    animation, otherwise the number is the number of frames to step through
    every decisecond.
  */
  public void setAnimation(int nSpeed) {
    nAnimationRate = nSpeed;
  }

  /**
    Sets the color of the grid.  The correct color is used when the piece is
    in the correct location.  The Incorrect color is used when the piece is not
    in the correct location.  The select color is used to indicate which piece
    is currently selected.
  */
  public void setColor(Color cCorrect, Color cIncorrect, Color cSelect) {
    colorCorrect = cCorrect;
    colorIncorrect = cIncorrect;
    colorSelect = cSelect;
  }

  /**
    Enables or disables the displaying of the grid.
  */
  public void setGrid(boolean state) {
    bGridVisible = state;
  }

  /**
    This solves the puzzle and causes the game to be won.  This is useful in
    the stand alone versions of the game where the puzzle must be shown before
    it is scrambled.
  */
  public void forceWin() {
    for (int cntr = 0; cntr < (nGridHeight * nGridWidth); ++cntr)
      nPuzzleData[cntr] = cntr;
    hasWon = true;
  }

  /**
    Swaps two pieces of the puzzle
  */
  public void swap(int x1, int y1, int x2, int y2) {
    int temp = nPuzzleData[y1 * nGridWidth + x1];
    nPuzzleData[y1 * nGridWidth + x1] = nPuzzleData[y2 * nGridWidth + x2];
    nPuzzleData[y2 * nGridWidth + x2] = temp;
    ++_counterValue;
  }

  /**
    Mixes up the puzzle.
  */
  public void scramble() {
    int nNumCells = nGridWidth * nGridHeight;
    int cntr, toswap, temp;
    Random r = new Random();

    for (cntr = 0; cntr < nNumCells; ++cntr) {
      toswap = Math.abs(r.nextInt()) % nNumCells;
      temp = nPuzzleData[cntr];
      nPuzzleData[cntr] = nPuzzleData[toswap];
      nPuzzleData[toswap] = temp;
    }
    hasWon = false;
    nCurrentSelection = -1;

    _counterValue = 0;
    _timerValue = 0;
    _percentValue = 0;
    _startTime = System.currentTimeMillis();
  }

  /**
    Checks to see if the game has been won.
        (also calculates current percentage)
  */
    public boolean isWon()
    {
        int nNumCells = nGridWidth * nGridHeight;
        int correctCount = 0;

        for (int cntr = 0; cntr < nNumCells; ++cntr)
        {
            if (nPuzzleData[cntr] == cntr)
                ++correctCount;
        }

        _percentValue = correctCount * 100 / nNumCells;
        if (correctCount == nNumCells)
            return true;
        else
            return false;
    }
    
  /**
    Paints the animation.
  */
  synchronized void paintAnimation(Graphics g) {
    nAnimationFrame += nAnimationRate;

    int cellSize = Math.min(nGridPixelsX, nGridPixelsY);

    if (nAnimationFrame >= cellSize) {
      animationPlaying = false;
      nAnimationFrame = cellSize;
    }

    int nImWidth = iPuzzle.getWidth(this);
    int nImHeight = iPuzzle.getHeight(this);
    int PicLeft = (getSize().width - nImWidth) / 2;
    int PicTop = (getSize().height -  nImHeight) / 2;
    if (_usePlayfield)
    {
        PicLeft = _playfieldRect.x;
        PicTop = _playfieldRect.y;
    }
    int nCell1X, nCell1Y, nPCell1X, nPCell1Y;
    int nCell2X, nCell2Y, nPCell2X, nPCell2Y;

    nCell1X = PicLeft + nPicLeftGap + (nCurrentSelection % nGridWidth) * nGridPixelsX;
    nCell1Y = PicTop + nPicTopGap + (nCurrentSelection / nGridWidth) * nGridPixelsY;
    nPCell1X = nPicLeftGap + (nPuzzleData[nCurrentSelection] % nGridWidth) * nGridPixelsX;
    nPCell1Y = nPicTopGap + (nPuzzleData[nCurrentSelection] / nGridWidth) * nGridPixelsY;

    nCell2X = PicLeft + nPicLeftGap + (nSecondSelection % nGridWidth) * nGridPixelsX;
    nCell2Y = PicTop + nPicTopGap + (nSecondSelection / nGridWidth) * nGridPixelsY;
    nPCell2X = nPicLeftGap + (nPuzzleData[nSecondSelection] % nGridWidth) * nGridPixelsX;
    nPCell2Y = nPicTopGap + (nPuzzleData[nSecondSelection] / nGridWidth) * nGridPixelsY;

    // Clear the animated areas
    g.setColor(Color.black);
    g.fillRect(nCell1X, nCell1Y, nGridPixelsX-1, nGridPixelsY-1);
    g.fillRect(nCell2X, nCell2Y, nGridPixelsX-1, nGridPixelsY-1);

    // determin which half of the swap we are in
    if (nAnimationFrame > (cellSize / 2)) {
      int temp = (cellSize / 2) - (nAnimationFrame - (cellSize / 2));
      g.drawImage(iPuzzle, nCell1X + temp, nCell1Y + temp,
            nCell1X + nGridPixelsX - temp, nCell1Y + nGridPixelsY - temp,
            nPCell1X, nPCell1Y,
            nPCell1X + nGridPixelsX, nPCell1Y + nGridPixelsY, this);
      g.drawImage(iPuzzle, nCell2X + temp, nCell2Y + temp,
            nCell2X + nGridPixelsX - temp, nCell2Y + nGridPixelsY - temp,
            nPCell2X, nPCell2Y,
            nPCell2X + nGridPixelsX, nPCell2Y + nGridPixelsY, this);
    }
    else {
      int temp = nAnimationFrame;
      g.drawImage(iPuzzle, nCell1X + temp, nCell1Y + temp,
            nCell1X + nGridPixelsX - temp, nCell1Y + nGridPixelsY - temp,
            nPCell2X, nPCell2Y,
            nPCell2X + nGridPixelsX, nPCell2Y + nGridPixelsY, this);
      g.drawImage(iPuzzle, nCell2X + temp, nCell2Y + temp,
            nCell2X + nGridPixelsX - temp, nCell2Y + nGridPixelsY - temp,
            nPCell1X, nPCell1Y,
            nPCell1X + nGridPixelsX, nPCell1Y + nGridPixelsY, this);
    }

    if (nAnimationFrame == cellSize)
        nCurrentSelection = -1;
  }


  /**
    paints the image and (optionally) the grid.
  */
	public void paint(Graphics g) {
    if (iPuzzle == null)
      return;

    if (animationPlaying) {
      paintAnimation(g);
      if (animationPlaying)
        return;
    }

    int nImWidth = iPuzzle.getWidth(this);
    int nImHeight = iPuzzle.getHeight(this);
    int PicLeft = (getSize().width - nImWidth) / 2;
    int PicTop = (getSize().height -  nImHeight) / 2;
    if (_usePlayfield)
    {
        PicLeft = _playfieldRect.x;
        PicTop = _playfieldRect.y;
    }
    int cntrX, cntrY, nCellX, nCellY, nPCellX, nPCellY, nCurCell;

    // draw the top and left borders of the image (if any)
    if (nPicTopGap > 0)
      g.drawImage(iPuzzle, PicLeft, PicTop, PicLeft+nImWidth, PicTop+nPicTopGap,
            0, 0, nImWidth, nPicTopGap, this);
    if (nPicLeftGap > 0)
      g.drawImage(iPuzzle, PicLeft, PicTop, PicLeft+nPicLeftGap, PicTop+nImHeight,
            0, 0, nPicLeftGap, nImHeight, this);

    // Draw the rest of the image
    for (cntrX = 0; cntrX < nGridWidth; ++cntrX)
      for (cntrY = 0; cntrY < nGridHeight; ++cntrY) {
        nCellX = PicLeft + nPicLeftGap + cntrX * nGridPixelsX;
        nCellY = PicTop + nPicTopGap + cntrY * nGridPixelsY;
        nCurCell = nPuzzleData[cntrX + cntrY * nGridWidth];
        nPCellX = nPicLeftGap + (nCurCell % nGridWidth) * nGridPixelsX;
        nPCellY = nPicTopGap + (nCurCell / nGridWidth) * nGridPixelsY;
        g.drawImage(iPuzzle, nCellX, nCellY,
              nCellX + nGridPixelsX, nCellY + nGridPixelsY,
              nPCellX, nPCellY,
              nPCellX + nGridPixelsX, nPCellY + nGridPixelsY, this);

        // draw the grid (if not solved and grid visibility is on)
        if ((!hasWon) && (bGridVisible)) {
          if ( nCurCell == (cntrX + cntrY * nGridWidth) )
            g.setColor(colorCorrect);
          else
            g.setColor(colorIncorrect);
          if ((cntrX + cntrY * nGridWidth) == nCurrentSelection)
            g.setColor(colorSelect);
          g.drawLine(nCellX, nCellY, nCellX, nCellY + nGridPixelsY-1);
          g.drawLine(nCellX, nCellY + nGridPixelsY-1, nCellX + nGridPixelsX-1, nCellY + nGridPixelsY-1);
          g.drawLine(nCellX + nGridPixelsX-1, nCellY + nGridPixelsY-1, nCellX + nGridPixelsX-1, nCellY);
          g.drawLine(nCellX + nGridPixelsX-1, nCellY, nCellX, nCellY);
        }  // if visiblity is off, draw just the selection (if any).
        else if ((!hasWon) && (!bGridVisible) && (nCurrentSelection >= 0)) {
          g.setColor(colorSelect);
          nCellX = PicLeft + nPicLeftGap + (nCurrentSelection % nGridWidth) * nGridPixelsX;
          nCellY = PicTop + nPicTopGap + (nCurrentSelection / nGridWidth) * nGridPixelsY;
          g.drawRect(nCellX, nCellY, nGridPixelsX-1, nGridPixelsY-1);
        }
      }

    // draw the bottom and right borders of the image (if any)
    if (nPicRightGap > 0)
      g.drawImage(iPuzzle,PicLeft + nImWidth - nPicRightGap,PicTop,
            PicLeft + nImWidth,PicTop + nImHeight,
            nImWidth - nPicRightGap,0, nImWidth, nImHeight, this);
    if (nPicBotGap > 0)
      g.drawImage(iPuzzle, PicLeft, PicTop + nImHeight - nPicBotGap,
            PicLeft+ nImWidth, PicTop + nImHeight,
            0, nImHeight - nPicBotGap, nImWidth, nImHeight, this);
    paintLabels(g);
	}


  /**
    Remove flicker from redraws by forcing update not to clear screen.
  */
	public void update(Graphics g) {
			paint(g);
	}

  /**
    Resets the screen to its starting condition.
  */
  public void reset() {
  }


  /**
    Decisecond tick notification.  The controlling object calles this function
    once every decisecond (10th of a second).  It can be used for handling
    animations and other timing dependancies.
  */
  public void tick() {
    if (animationPlaying)
      repaint();
  }


  /**
    Called whenever the mouse is moved so that the screen can handle it if
    it needs to.
  */
  public void mouse_moved(int x, int y) {}

  /**
    Handle the pressing of the mouse button.  Returns a non-zero value if this
    results in an external action (such as a switch to another screen). The
    external action values are described in the MWAGConst static class.
  */
  public int mouse_released(int x, int y) {
    if (animationPlaying)
      return 0;

    int gx, gy;
    int nImWidth = iPuzzle.getWidth(this);
    int nImHeight = iPuzzle.getHeight(this);
    int PicLeft = (getSize().width - nImWidth) / 2;
    int PicTop = (getSize().height -  nImHeight) / 2;
    if (_usePlayfield)
    {
        PicLeft = _playfieldRect.x;
        PicTop = _playfieldRect.y;
    }
    
    gx = (x - PicLeft - nPicLeftGap) / nGridPixelsX;
    gy = (y - PicTop - nPicTopGap) / nGridPixelsY;

    if ( (gx < 0) || (gx >= nGridWidth) || (gy < 0) || (gy >= nGridHeight) )
      return 0;
    if (hasWon)
      return PuzScreen.SCREEN_SELECTION;

    if (nCurrentSelection == -1)
      nCurrentSelection = gy * nGridWidth + gx;
    else {
      swap(nCurrentSelection % nGridWidth,nCurrentSelection / nGridWidth, gx,gy);
      hasWon = isWon();
      nSecondSelection = gy * nGridWidth + gx;
      if (nAnimationRate > 0) {
        nAnimationFrame = 0;
        animationPlaying = true;
      }
      else
        nCurrentSelection = -1;
    }
    repaint();
    return 0;
  }

// ************************************************
// ***** METHODS - Label setting and handling *****
// ************************************************

    /**
     * For compatibility with PPA, the playfield will only be used if this method is called.
     * If this method is not called, the full screen is the viewport.
     *
     * @param rect - where to place the counter
     */
    public void setPlayfield(Rectangle rect)
    {
        _playfieldRect = rect;
        _usePlayfield = true;
    }
    
    /**
     * For compatibility with PPA, the Moves counter will only be used if this method is called.
     *
     * @param rect - where to place the counter
     */
    public void setCounter(Rectangle rect)
    {
        _counterRect = rect;
        _counterVisible = true;
    }

    /**
    * For compatibility with PPA, the Time counter will only be used if this method is called.
     *
     * @param rect - where to place the counter
     */
    public void setTimer(Rectangle rect)
    {
        _timerRect = rect;
        _timerVisible = true;
    }

    /**
     * For compatibility with PPA, the Percent done will only be used if this method is called.
     *
     * @param rect - where to place the counter
     */
    public void setPercent(Rectangle rect)
    {
        _percentRect = rect;
        _percentVisible = true;
    }

    /**
     * Draw the labels. Only drawn if label set using above methods.
     *
     * @param g graphics context
    */
    public void paintLabels(Graphics g)
    {
        // We don't want to set the font needlessly, so we quickly see if we
        // even need to run this method before doing individual checks
        if ((_counterVisible | _timerVisible | _percentVisible) == false)
            return;

        int center;
        String content;
        Font oldFont = g.getFont();
        g.setFont(LABEL_FONT);
        FontMetrics fm = g.getFontMetrics();
        if (_counterVisible)
        {
            g.setColor(Color.white);
            g.fillRect(_counterRect.x, _counterRect.y, _counterRect.width, _counterRect.height);
            g.setColor(Color.black);
            center = (_counterRect.width - fm.stringWidth(MOVES_LABEL)) / 2;
            g.drawString(MOVES_LABEL, _counterRect.x + center, _counterRect.y + (_counterRect.height/2));
            content = "" + _counterValue;
            center = (_counterRect.width - fm.stringWidth(content)) / 2;
            g.drawString(content, _counterRect.x + center, _counterRect.y + _counterRect.height);
        }
        
        if (_timerVisible)
        {
            g.setColor(Color.white);
            g.fillRect(_timerRect.x, _timerRect.y, _timerRect.width, _timerRect.height);
            g.setColor(Color.black);
            center = (_timerRect.width - fm.stringWidth(TIME_LABEL)) / 2;
            g.drawString(TIME_LABEL, _timerRect.x + center, _timerRect.y + (_timerRect.height/2));
            long curTime = System.currentTimeMillis();
            long timeDif = (curTime - _startTime) / 1000;
            if ( (timeDif % 60) < 10)
                content = (timeDif / 60) + ":0" + (timeDif % 60);
            else
                content = (timeDif / 60) + ":" + (timeDif % 60);

            center = (_timerRect.width - fm.stringWidth(content)) / 2;
            g.drawString(content, _timerRect.x + center, _timerRect.y + _timerRect.height);
            
        }

        if (_percentVisible)
        {
            g.setColor(Color.white);
            g.fillRect(_percentRect.x, _percentRect.y, _percentRect.width, _percentRect.height);
            g.setColor(Color.black);
            center = (_percentRect.width - fm.stringWidth(DONE_LABEL)) / 2;
            g.drawString(DONE_LABEL, _percentRect.x + center, _percentRect.y + (_percentRect.height/2));
            content = "" + _percentValue;
            center = (_percentRect.width - fm.stringWidth(content)) / 2;
            g.drawString(content, _percentRect.x + center, _percentRect.y + _percentRect.height);            
        }
        g.setFont(oldFont);
    }
}


